/* USAGE: mmail2mbox inputdir outputdir
 *
 * Given the directory where mail is stored for a Gnus mmail file as the
 * inputdir (i.e., there are the drafts and mail folders directly under
 * the directory), mmail2mbox will convert all the mail folders available
 * for reading into mbox's and store the mbox files under the outputdir
 * directory.
 */

/*Copyright (c) 2005, Aaron Hsu
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 * * Neither the name of the Wen-yao Animation & Programing nor the names
 *   of its contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define SYNTAXSTRG "mmail2mbox <input directory> <output directory>\n"

#define DC "\\"

#include <io.h>
#include <stdio.h>
#include <direct.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

struct date_format {
	int day;
	char month[4];
	int year;
	char weekday[4];
	char time[15];
};

struct overviewentry {
	int index;
	size_t es_max;
	struct date_format date;
	char *es;
};

size_t apnd_fname(char **fname, char *s1, char *d, char *s2)
{
	size_t max;
	
	max = strlen(s1) + strlen(s2) + strlen(d) + 1;
	*fname = realloc(fname, max * sizeof(char));
	if (*fname == NULL) {
		fprintf(stderr, "apnd_fname(): Insufficient Memory.\n");
		errno = 1;
		return 0;
	}
	strncpy(*fname, s1, max);
	strncat(*fname, d, max - strlen(*fname) - 1);
	strncat(*fname, s2, max - strlen(*fname) - 1);
	
	return max;
}

/* Fills es with the e-mail address found in s, returning the max space
 * as the number of characters
 */
 
int putes(char **es, char *s, int es_max)
{
	int i; /* Loop index */
	int f; /* Flag for determining start and stops in extractions */
	int u; /* "Usage" index for es string. */
	f = 0;
	u = 0;
	
	for (i = 0; s[i] != '\0'; i++) { /* Test ".*" <.*> format first */
		if (f == 0) { /* Have not found first '<' */
			if (s[i] == '<') f = 1; /* if found, set flag */
		} else {
			if (es_max <= u + 1) { /* Need more memory */
				es_max = u + 50; /* Set size relative to usage */
				*es = realloc(*es, (u + 50) * sizeof(char));
				if (*es == NULL) {
					fprintf(stderr, "putes(): Insufficient memory.\n");
					errno = 1;
					return 0;
				}
			}
			if (s[i] != '>') { /* continue extraction */
				(*es)[u++] = s[i];
			} else {
				(*es)[u] = '\0'; /* Complete string */
				return es_max;
			}
		}
	}
	for (i = 0; s[i] != '\0'; i++) { /* Just extract address if no special formatting */
		if (es_max <= u + 1) {
			es_max = u + 50;
			*es = realloc(*es, (u + 50) * sizeof(char));
			if (*es == NULL) {
				fprintf(stderr, "putes(): Insufficient memory.\n");
				errno = 3;
				return 0;
			}
		}
		(*es)[u++] = s[i];
	}
	if (u > 0) { /* Something was extracted */
		(*es)[u] = '\0';
		return es_max;
	}
	fprintf(stderr, "putes(): Failed to complete extraction.\n");
	errno = 2;
	return 0;
}

/* Takes an overview entry and fills it with the data c according to
 * the level indicated by i.
 */
 
struct overviewentry *fillovrntry(struct overviewentry **ovrntry, char *s, int i)
{
	if (*ovrntry == NULL) { /* No memory allocated for entry */
		*ovrntry = malloc(sizeof(struct overviewentry));
		(*ovrntry)->es = malloc(sizeof(char) * 50);
		(*ovrntry)->es_max = 50;
		(*ovrntry)->es[0] = '\0';
	}
	if (*ovrntry == NULL || (*ovrntry)->es == NULL) {
		fprintf(stderr, "fillovrntry(): Insufficient memory.\n");
		errno = 1;
		return NULL;
	}
	switch (i) { /* Examine levels and fill accordingly */
		case 0: (*ovrntry)->index = atoi(s); break;
		case 2: /* Extraction of e-mail address */
			(*ovrntry)->es_max = putes(&(*ovrntry)->es, s, (*ovrntry)->es_max);
			if ((*ovrntry)->es_max == 0) {
				fprintf(stderr, "fillovrntry(): Error %d, failed to extract envsender.\n", errno);
				errno = 2;
				return NULL;
			}
			break;
		case 3: /* Extraction of Date */
			sscanf(s,
				   "%3s, %d %3s %4d %8s %*s",
				   (*ovrntry)->date.weekday,
				   &((*ovrntry)->date.day),
				   (*ovrntry)->date.month,
				   &((*ovrntry)->date.year),
				   (*ovrntry)->date.time);
			break;
	}
	return *ovrntry;
}

/* Reads an overview file and obtains the next overview entry from that
 * file.
 */
 
struct overviewentry *getovrntry(struct overviewentry **ovrntry, FILE *overview)
{
	char c;
	char *s; /* String to store the contents of one field */
	int m; /* Max allocation units available */
	int u; /* Used allocation units */
	int i; /* Used for index level/field identification */
	c = fgetc(overview);
	i = 0;
	m = 0;
	u = 0;
	s = NULL;

	if (feof(overview) != 0) { /* End of file, free memory */
		free((*ovrntry)->es);
		free(*ovrntry);
		errno = 0;
		return NULL;
	}
	
	/* The basic idea is looping until the entire line in an overview 
	 * entry has been read. Each tab that is encountered is discarded 
	 * and the characters read up to that point since the last tab 
	 * are sent to the fillovrntry() function to populate the 
	 * overview struct accordingly.
	 */
	while (!feof(overview) && c != '\n') {
		if (c == '\t') {
			*ovrntry = fillovrntry(ovrntry, s, i);
			if (*ovrntry == NULL) {
				fprintf(stderr, "getovrntry(): Error %d, failed to populate overview entry structure.\n", errno);
				errno = 2;
				return NULL;
			}
			u = 0;
			i++;
		} else {
			if (m <= u + 1) { /* Need more memory */
				m = (m * 2) + 2; /* + 2 ensures space should m == 0 */
				s = realloc(s, m * sizeof(char));
				if (s == NULL) {
					fprintf(stderr, "getovrntry(): Insufficient Memory.\n");
					errno = 1;
					return NULL;
				}
				continue;
			}
			s[u++] = c;
			s[u] = '\0';
		}
		c = fgetc(overview);
	}
	free(s);
	return *ovrntry;
}

int writeovrntry(FILE *ofile, struct overviewentry *ovrntry)
{
	char c;
	char buf[50]; /* Spare buffer large enough to hold more digits than long */
	FILE *ifile;
	
	if (fputs("From ", ofile) == EOF) { /* Write first part of mbox entry */
		fprintf(stderr, "writeovrntry(): Failed to write file.\n");
		errno = 1;
		return EXIT_FAILURE;
	}
	if (fputs(ovrntry->es, ofile) == EOF) { /* Add email to entry line */
		fprintf(stderr, "writeovrntry(): Failed to write file.\n");
		errno = 2;
		return EXIT_FAILURE;
	}
	fprintf(ofile, /* put date in as final part of entry line */
			" %s %s %2d %s %4d\n",
			ovrntry->date.weekday,
			ovrntry->date.month,
			ovrntry->date.day,
			ovrntry->date.time,
			ovrntry->date.year);
	ifile = fopen(ltoa(ovrntry->index, buf, 10), "r");
	if (ifile == NULL) {
		fprintf(stderr, "writeovrntry(): Error %d, Failed to open '%s'\n", errno, buf);
		errno = 3;
		return EXIT_FAILURE;
	}
	c = fgetc(ifile);
	while (!feof(ifile)) { /* Write email message to file */
		if (fputc(c, ofile) == EOF) {
			fprintf(stderr, "writeovrntry(): Failed writing file.\n");
			errno = 4;
			return EXIT_FAILURE;
		}
		c = fgetc(ifile);
	}
	if (fputc('\n', ofile) == EOF) { /* Add final line after rest of message */
		fprintf(stderr, "writeovrntry(): Failed writing file.\n");
		errno = 5;
		return EXIT_FAILURE;
	}
	fclose(ifile);
	return EXIT_SUCCESS;
}

/* Reads the overview file from inputdir and generates the appropriate
 * mbox file with the name of the inputdir filename in the outputdir
 */
 
int gen_mbox(char *mail_dir, char *o_dir, char *m_name)
{
	char *m_dir; /* Used to set the directory of the current mbox */
	char *o_name; /* Used to hold the output filename */
	size_t o_name_max; /* Allocated space for o_name */
	size_t m_dir_max; /* Allocated space for m_dir */
	struct overviewentry *ovrntry;
	FILE *ofile;
	FILE *overview;

	ovrntry = NULL;

	/* Check output directory and verify existance */
	if (access(o_dir, F_OK)) {
		if(mkdir(o_dir)) {
			fprintf(stderr, "gen_mbox(): Failed to create directory: %s.\n", o_dir);
			errno = 5;
			return EXIT_FAILURE;
		}
	}
	
	/* Open output file */
	o_name_max = apnd_fname(&o_name, o_dir, DC, m_name);
	if (o_name_max == 0) {
		fprintf(stderr, "gen_mbox(): Error %d, Failed to composite filename.\n", errno);
		errno = 6;
		return EXIT_FAILURE;
	}
	ofile = fopen(o_name, "a");
	if (ofile == NULL) {
		fprintf(stderr, "gen_mbox(): Error %d, failed to open output file: %s\n", errno, o_name);
		errno = 2;
		return EXIT_FAILURE;
	}

	/* Change directory to mail_dir + m_name */
	m_dir_max = apnd_fname(&m_dir, mail_dir, DC, m_name);
	if (m_dir_max == 0) {
		fprintf(stderr, "gen_mbox(): Error %d, Failed to composite filename.\n", errno);
		errno = 7;
		return EXIT_FAILURE;
	}
	if (chdir(m_dir)) {
		fprintf(stderr, "gen_mbox(): Error %d, failed to change to directory: %s\n", errno, m_dir);
		errno = 1;
		return EXIT_FAILURE;
	}

	/* Open .overview file */
	overview = fopen(".overview", "r");
	if (overview == NULL) {
		fprintf(stderr, "gen_mbox(): Error %d, failed to open overview file: %s%s.overview\n", errno, m_dir, DC);
		errno = 3;
		return EXIT_FAILURE;
	}
	
	/* LOOP Read .overview entry */
	ovrntry = getovrntry(&ovrntry, overview);
	while (ovrntry != NULL) { 
		/* Write .overview entry */
		if (writeovrntry(ofile, ovrntry) == EXIT_FAILURE) {
			fprintf(stderr, "gen_mbox(): Error %d, failed to write overview entry.\n", errno);
			errno = 4;
			return EXIT_FAILURE;
		}
		ovrntry = getovrntry(&ovrntry, overview);
	}
	if (errno == 0 && ovrntry == NULL) {
		fclose(overview);
		fclose(ofile);
		free(m_dir);
		free(o_name);
		return EXIT_SUCCESS;
	}
	fprintf(stderr, "gen_mbox(): Error %d, failed to complete parsing of overview file.\n", errno);
	fclose(overview);
	fclose(ofile);
	return EXIT_FAILURE;
}

int main(int argc, char *argv[])
{
	DIR *inputdir;
	struct dirent *mboxdir;
	char *maildirstrg;
	char *origcwd;
	size_t maildirstrgmax;

	origcwd = getcwd(NULL, 0);

	/* Check Args */
	if (argc != 3) {
		fprintf(stderr, "main(): Invalid number of arguments.\n");
		printf("Syntax: %s", SYNTAXSTRG);
		return EXIT_FAILURE;
	}
	
	/* Open Maildir/mail directory */
	maildirstrgmax = apnd_fname(&maildirstrg, argv[1], DC, "mail");	
	if (maildirstrgmax == 0) {
		fprintf(stderr, "main(): Error %d, Failed to composite filename.\n", errno);
		return EXIT_FAILURE;
	}
	inputdir = opendir(maildirstrg);
	if (inputdir == NULL) {
		fprintf(stderr, "main(): Error %d, failed to open '%s'.\n", errno, maildirstrg);
		return EXIT_FAILURE;
	}
	
	/* LOOP write MBOX files */
	mboxdir = readdir(inputdir);
	while (mboxdir != NULL) {
		if (!strncmp(".", mboxdir->d_name, 1) || !strncmp("..", mboxdir->d_name, 2)) {
			mboxdir = readdir(inputdir);
			continue;
		}
		if (gen_mbox(maildirstrg, argv[2], mboxdir->d_name) == EXIT_FAILURE) {
			fprintf(stderr, "main(): Error %d, failed to generate mbox file from: %s%s%s.\n", errno, maildirstrg, DC, mboxdir->d_name);
			return EXIT_FAILURE;
		}
		chdir(origcwd);
		mboxdir = readdir(inputdir);
	}

	/* Switch to original working directory */
	if (chdir(origcwd) != 0) {
		fprintf(stderr, "main(): Error %d, failed to change to original directory '%s'\n", errno, origcwd);
		return EXIT_FAILURE;
	}

	/* Clear Memory */
	free(maildirstrg);
	free(origcwd);
	closedir(inputdir);

	return EXIT_SUCCESS;
}
